@@ -11,9 +11,11 @@
 
 #include <string.h>
 
+#include <gio/gnetworking.h>
+
 #include "soup-address.h"
 #include "soup.h"
-#include "soup-marshal.h"
+#include "soup-misc-private.h"
 
 /**
  * SECTION:soup-address
@@ -22,6 +24,10 @@
  * #SoupAddress represents the address of a TCP connection endpoint:
  * both the IP address and the port. (It is somewhat like an
  * object-oriented version of struct sockaddr.)
+ *
+ * Although #SoupAddress is still used in some libsoup API's, it
+ * should not be used in new code; use GLib's #GNetworkAddress or
+ * #GSocketAddress instead.
  **/
 
 enum {
@@ -46,9 +52,7 @@ typedef struct {
 	const char *protocol;
 
 	GMutex lock;
-	GSList *async_lookups;
 } SoupAddressPrivate;
-#define SOUP_ADDRESS_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_ADDRESS, SoupAddressPrivate))
 
 /* sockaddr generic macros */
 #define SOUP_SIN(priv) ((struct sockaddr_in *)priv->sockaddr)
@@ -93,13 +97,14 @@ typedef struct {
 static void soup_address_connectable_iface_init (GSocketConnectableIface *connectable_iface);
 
 G_DEFINE_TYPE_WITH_CODE (SoupAddress, soup_address, G_TYPE_OBJECT,
+                         G_ADD_PRIVATE (SoupAddress)
 			 G_IMPLEMENT_INTERFACE (G_TYPE_SOCKET_CONNECTABLE,
 						soup_address_connectable_iface_init))
 
 static void
 soup_address_init (SoupAddress *addr)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 
 	g_mutex_init (&priv->lock);
 }
@@ -108,7 +113,7 @@ static void
 soup_address_finalize (GObject *object)
 {
 	SoupAddress *addr = SOUP_ADDRESS (object);
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 
 	g_free (priv->sockaddr);
 	g_free (priv->name);
@@ -131,7 +136,7 @@ soup_address_constructor (GType                  type,
 		type, n_construct_properties, construct_properties);
 	if (!addr)
 		return NULL;
-	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	priv = soup_address_get_instance_private (SOUP_ADDRESS (addr));
 
 	if (!priv->name && !priv->sockaddr) {
 		g_object_unref (addr);
@@ -145,7 +150,7 @@ static void
 soup_address_set_property (GObject *object, guint prop_id,
 			   const GValue *value, GParamSpec *pspec)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (object);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (SOUP_ADDRESS (object));
 	SoupAddressFamily family;
 	struct sockaddr *sa;
 	int len, port;
@@ -209,7 +214,7 @@ static void
 soup_address_get_property (GObject *object, guint prop_id,
 			   GValue *value, GParamSpec *pspec)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (object);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (SOUP_ADDRESS (object));
 
 	switch (prop_id) {
 	case PROP_NAME:
@@ -244,8 +249,6 @@ soup_address_class_init (SoupAddressClass *address_class)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (address_class);
 
-	g_type_class_add_private (address_class, sizeof (SoupAddressPrivate));
-
 	/* virtual method override */
 	object_class->constructor  = soup_address_constructor;
 	object_class->finalize     = soup_address_finalize;
@@ -365,7 +368,7 @@ soup_address_new (const char *name, guint port)
  * Returns a #SoupAddress equivalent to @sa (or %NULL if @sa's
  * address family isn't supported)
  *
- * Return value: (allow-none): the new #SoupAddress
+ * Return value: (nullable): the new #SoupAddress
  **/
 SoupAddress *
 soup_address_new_from_sockaddr (struct sockaddr *sa, int len)
@@ -379,6 +382,17 @@ soup_address_new_from_sockaddr (struct sockaddr *sa, int len)
 			     NULL);
 }
 
+SoupAddress *
+soup_address_new_from_gsockaddr (GSocketAddress *addr)
+{
+	struct sockaddr_storage sa;
+
+	g_socket_address_to_native (addr, &sa, sizeof (sa), NULL);
+	return g_object_new (SOUP_TYPE_ADDRESS,
+			     SOUP_ADDRESS_SOCKADDR, &sa,
+			     NULL);
+}
+
 /**
  * SoupAddressFamily:
  * @SOUP_ADDRESS_FAMILY_INVALID: an invalid %SoupAddress
@@ -404,7 +418,7 @@ soup_address_new_from_sockaddr (struct sockaddr *sa, int len)
  * for @family (or %NULL if @family isn't supported), suitable for
  * using as a listening #SoupSocket.
  *
- * Return value: (allow-none): the new #SoupAddress
+ * Return value: (nullable): the new #SoupAddress
  **/
 SoupAddress *
 soup_address_new_any (SoupAddressFamily family, guint port)
@@ -429,14 +443,17 @@ soup_address_new_any (SoupAddressFamily family, guint port)
  * soup_address_is_resolved() to safely test whether or not an address
  * is resolved before fetching its name or address.
  *
- * Return value: (allow-none): the hostname, or %NULL if it is not known.
+ * Return value: (nullable): the hostname, or %NULL if it is not known.
  **/
 const char *
 soup_address_get_name (SoupAddress *addr)
 {
-	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), NULL);
+	SoupAddressPrivate *priv;
 
-	return SOUP_ADDRESS_GET_PRIVATE (addr)->name;
+	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), NULL);
+	priv = soup_address_get_instance_private (addr);
+
+	return priv->name;
 }
 
 /**
@@ -452,7 +469,7 @@ soup_address_get_name (SoupAddress *addr)
  * soup_address_is_resolved() to safely test whether or not an address
  * is resolved before fetching its name or address.
  *
- * Return value: (allow-none) (transfer none): the sockaddr, or %NULL
+ * Return value: (nullable) (transfer none): the sockaddr, or %NULL
  **/
 struct sockaddr *
 soup_address_get_sockaddr (SoupAddress *addr, int *len)
@@ -460,7 +477,7 @@ soup_address_get_sockaddr (SoupAddress *addr, int *len)
 	SoupAddressPrivate *priv;
 
 	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), NULL);
-	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	priv = soup_address_get_instance_private (addr);
 
 	if (priv->sockaddr && len)
 		*len = SOUP_ADDRESS_FAMILY_SOCKADDR_SIZE (SOUP_ADDRESS_GET_FAMILY (priv));
@@ -481,7 +498,7 @@ soup_address_get_sockaddr (SoupAddress *addr, int *len)
 GSocketAddress *
 soup_address_get_gsockaddr (SoupAddress *addr)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 
 	return g_socket_address_new_from_native (priv->sockaddr,
 						 SOUP_ADDRESS_FAMILY_SOCKADDR_SIZE (SOUP_ADDRESS_GET_FAMILY (priv)));
@@ -490,7 +507,7 @@ soup_address_get_gsockaddr (SoupAddress *addr)
 static GInetAddress *
 soup_address_make_inet_address (SoupAddress *addr)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 	GSocketAddress *gsa;
 	GInetAddress *gia;
 
@@ -514,7 +531,7 @@ soup_address_make_inet_address (SoupAddress *addr)
  * soup_address_is_resolved() to safely test whether or not an address
  * is resolved before fetching its name or address.
  *
- * Return value: (allow-none): the physical address, or %NULL
+ * Return value: (nullable): the physical address, or %NULL
  **/
 const char *
 soup_address_get_physical (SoupAddress *addr)
@@ -522,7 +539,7 @@ soup_address_get_physical (SoupAddress *addr)
 	SoupAddressPrivate *priv;
 
 	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), NULL);
-	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	priv = soup_address_get_instance_private (addr);
 
 	if (!priv->sockaddr)
 		return NULL;
@@ -549,16 +566,72 @@ soup_address_get_physical (SoupAddress *addr)
 guint
 soup_address_get_port (SoupAddress *addr)
 {
-	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), 0);
+	SoupAddressPrivate *priv;
 
-	return SOUP_ADDRESS_GET_PRIVATE (addr)->port;
+	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), 0);
+	priv = soup_address_get_instance_private (addr);
+
+	return priv->port;
+}
+
+
+/* Tries to resolve priv->name as an IP address, possibly including an
+ * IPv6 scope id.
+ */
+static void
+maybe_resolve_ip (SoupAddress *addr)
+{
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
+	const char *pct, *ip;
+	char *tmp = NULL;
+	GSocketConnectable *gaddr;
+	GSocketAddressEnumerator *sa_enum;
+	GSocketAddress *saddr;
+
+	if (priv->sockaddr || !priv->name)
+		return;
+
+	pct = strchr (priv->name, '%');
+	if (pct)
+		ip = tmp = g_strndup (priv->name, pct - priv->name);
+	else
+		ip = priv->name;
+
+	if (!g_hostname_is_ip_address (ip)) {
+		g_free (tmp);
+		return;
+	}
+	g_free (tmp);
+
+	gaddr = g_network_address_new (priv->name, priv->port);
+	if (!gaddr)
+		return;
+
+	sa_enum = g_socket_connectable_enumerate (gaddr);
+	saddr = g_socket_address_enumerator_next (sa_enum, NULL, NULL);
+	if (saddr) {
+		priv->n_addrs = 1;
+		priv->sockaddr = g_new (struct sockaddr_storage, 1);
+		if (!g_socket_address_to_native (saddr, priv->sockaddr,
+						 sizeof (struct sockaddr_storage),
+						 NULL)) {
+			/* can't happen: We know the address format is supported
+			 * and the buffer is large enough
+			 */
+			g_warn_if_reached ();
+		}
+		g_object_unref (saddr);
+	}
+
+	g_object_unref (sa_enum);
+	g_object_unref (gaddr);
 }
 
 
 static guint
 update_addrs (SoupAddress *addr, GList *addrs, GError *error)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 	GInetAddress *gia;
 	GSocketAddress *gsa;
 	int i;
@@ -597,7 +670,7 @@ update_addrs (SoupAddress *addr, GList *addrs, GError *error)
 static guint
 update_name (SoupAddress *addr, const char *name, GError *error)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 
 	if (error) {
 		if (error->domain == G_IO_ERROR &&
@@ -615,56 +688,48 @@ update_name (SoupAddress *addr, const char *name, GError *error)
 }
 
 typedef struct {
+	SoupAddress         *addr;
 	SoupAddressCallback  callback;
 	gpointer             callback_data;
 } SoupAddressResolveAsyncData;
 
 static void
-complete_resolve_async (SoupAddress *addr, guint status)
+complete_resolve_async (SoupAddressResolveAsyncData *res_data, guint status)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
-	SoupAddressResolveAsyncData *res_data;
-	GSList *lookups, *l;
 	GSource *current_source;
 	GMainContext *current_context;
 
-	lookups = priv->async_lookups;
-	priv->async_lookups = NULL;
+	if (res_data->callback) {
+		/* Awful hack; to make soup_socket_connect_async()
+		 * with an non-default async_context work correctly,
+		 * we need to ensure that the non-default context
+		 * (which we're now running in) is the thread-default
+		 * when the callbacks are run...
+		 */
+		current_source = g_main_current_source ();
+		if (current_source && !g_source_is_destroyed (current_source))
+			current_context = g_source_get_context (current_source);
+		else
+			current_context = NULL;
+		g_main_context_push_thread_default (current_context);
 
-	/* Awful hack; to make soup_socket_connect_async() with an
-	 * non-default async_context work correctly, we need to ensure
-	 * that the non-default context (which we're now running in)
-	 * is the thread-default when the callbacks are run...
-	 */
-	current_source = g_main_current_source ();
-	if (current_source && !g_source_is_destroyed (current_source))
-		current_context = g_source_get_context (current_source);
-	else
-		current_context = NULL;
-	g_main_context_push_thread_default (current_context);
+		res_data->callback (res_data->addr, status,
+				    res_data->callback_data);
 
-	for (l = lookups; l; l = l->next) {
-		res_data = l->data;
-
-		if (res_data->callback) {
-			res_data->callback (addr, status,
-					    res_data->callback_data);
-		}
-		g_slice_free (SoupAddressResolveAsyncData, res_data);
+		g_main_context_pop_thread_default (current_context);
 	}
-	g_slist_free (lookups);
 
-	g_main_context_pop_thread_default (current_context);
-
-	g_object_unref (addr);
+	g_object_unref (res_data->addr);
+	g_slice_free (SoupAddressResolveAsyncData, res_data);
 }
 
 static void
 lookup_resolved (GObject *source, GAsyncResult *result, gpointer user_data)
 {
 	GResolver *resolver = G_RESOLVER (source);
-	SoupAddress *addr = user_data;
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressResolveAsyncData *res_data = user_data;
+	SoupAddress *addr = res_data->addr;
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 	GError *error = NULL;
 	guint status;
 
@@ -689,7 +754,7 @@ lookup_resolved (GObject *source, GAsyncResult *result, gpointer user_data)
 	g_object_ref (addr);
 	g_object_set_data (G_OBJECT (addr), "async-resolved-error", error);
 
-	complete_resolve_async (addr, status);
+	complete_resolve_async (res_data, status);
 
 	g_object_set_data (G_OBJECT (addr), "async-resolved-error", NULL);
 	g_object_unref (addr);
@@ -698,9 +763,9 @@ lookup_resolved (GObject *source, GAsyncResult *result, gpointer user_data)
 }
 
 static gboolean
-idle_complete_resolve (gpointer addr)
+idle_complete_resolve (gpointer res_data)
 {
-	complete_resolve_async (addr, SOUP_STATUS_OK);
+	complete_resolve_async (res_data, SOUP_STATUS_OK);
 	return FALSE;
 }
 
@@ -746,63 +811,58 @@ soup_address_resolve_async (SoupAddress *addr, GMainContext *async_context,
 	SoupAddressPrivate *priv;
 	SoupAddressResolveAsyncData *res_data;
 	GResolver *resolver;
-	gboolean already_started;
 
 	g_return_if_fail (SOUP_IS_ADDRESS (addr));
-	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	priv = soup_address_get_instance_private (addr);
 	g_return_if_fail (priv->name || priv->sockaddr);
 
 	/* We don't need to do locking here because the async case is
 	 * not intended to be thread-safe.
 	 */
 
+	if (priv->name && !priv->sockaddr)
+		maybe_resolve_ip (addr);
 	if (priv->name && priv->sockaddr && !callback)
 		return;
 
 	res_data = g_slice_new0 (SoupAddressResolveAsyncData);
+	res_data->addr = g_object_ref (addr);
 	res_data->callback = callback;
 	res_data->callback_data = user_data;
 
-	already_started = priv->async_lookups != NULL;
-	priv->async_lookups = g_slist_prepend (priv->async_lookups, res_data);
-
-	if (already_started)
-		return;
-
-	g_object_ref (addr);
-
-	if (priv->name && priv->sockaddr) {
-		soup_add_completion (async_context, idle_complete_resolve, addr);
-		return;
-	}
-
-	resolver = g_resolver_get_default ();
 	if (async_context)
 		g_main_context_push_thread_default (async_context);
 
-	if (priv->name) {
-		g_resolver_lookup_by_name_async (resolver, priv->name,
-						 cancellable,
-						 lookup_resolved, addr);
-	} else {
-		GInetAddress *gia;
+	if (priv->name && priv->sockaddr)
+		soup_add_completion (async_context, idle_complete_resolve, res_data);
+	else {
+		resolver = g_resolver_get_default ();
 
-		gia = soup_address_make_inet_address (addr);
-		g_resolver_lookup_by_address_async (resolver, gia,
-						    cancellable,
-						    lookup_resolved, addr);
-		g_object_unref (gia);
+		if (priv->name) {
+			g_resolver_lookup_by_name_async (resolver, priv->name,
+							 cancellable,
+							 lookup_resolved, res_data);
+		} else {
+			GInetAddress *gia;
+
+			gia = soup_address_make_inet_address (addr);
+			g_resolver_lookup_by_address_async (resolver, gia,
+							    cancellable,
+							    lookup_resolved, res_data);
+			g_object_unref (gia);
+		}
+
+		g_object_unref (resolver);
 	}
 
 	if (async_context)
 		g_main_context_pop_thread_default (async_context);
-	g_object_unref (resolver);
 }
 
 static guint
 resolve_sync_internal (SoupAddress *addr, GCancellable *cancellable, GError **error)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 	GResolver *resolver;
 	guint status;
 	GError *my_err = NULL;
@@ -816,6 +876,10 @@ resolve_sync_internal (SoupAddress *addr, GCancellable *cancellable, GError **er
 	 * blocking op, and then re-lock it to modify @addr.
 	 */
 	g_mutex_lock (&priv->lock);
+
+	if (priv->name && !priv->sockaddr)
+		maybe_resolve_ip (addr);
+
 	if (!priv->sockaddr) {
 		GList *addrs;
 
@@ -841,6 +905,7 @@ resolve_sync_internal (SoupAddress *addr, GCancellable *cancellable, GError **er
 		g_free (name);
 	} else
 		status = SOUP_STATUS_OK;
+
 	g_mutex_unlock (&priv->lock);
 
 	if (my_err)
@@ -876,7 +941,7 @@ soup_address_resolve_sync (SoupAddress *addr, GCancellable *cancellable)
 	SoupAddressPrivate *priv;
 
 	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), SOUP_STATUS_MALFORMED);
-	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	priv = soup_address_get_instance_private (addr);
 	g_return_val_if_fail (priv->name || priv->sockaddr, SOUP_STATUS_MALFORMED);
 
 	return resolve_sync_internal (addr, cancellable, NULL);
@@ -899,7 +964,7 @@ soup_address_is_resolved (SoupAddress *addr)
 	gboolean resolved;
 
 	g_return_val_if_fail (SOUP_IS_ADDRESS (addr), FALSE);
-	priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	priv = soup_address_get_instance_private (addr);
 
 	g_mutex_lock (&priv->lock);
 	resolved = priv->sockaddr && priv->name;
@@ -922,7 +987,7 @@ soup_address_is_resolved (SoupAddress *addr)
 guint
 soup_address_hash_by_name (gconstpointer addr)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (SOUP_ADDRESS (addr));
 
 	g_return_val_if_fail (priv->name != NULL, 0);
 	return g_str_hash (priv->name);
@@ -963,8 +1028,8 @@ soup_address_hash_by_name (gconstpointer addr)
 gboolean
 soup_address_equal_by_name (gconstpointer addr1, gconstpointer addr2)
 {
-	SoupAddressPrivate *priv1 = SOUP_ADDRESS_GET_PRIVATE (addr1);
-	SoupAddressPrivate *priv2 = SOUP_ADDRESS_GET_PRIVATE (addr2);
+	SoupAddressPrivate *priv1 = soup_address_get_instance_private (SOUP_ADDRESS (addr1));
+	SoupAddressPrivate *priv2 = soup_address_get_instance_private (SOUP_ADDRESS (addr2));
 
 	g_return_val_if_fail (priv1->name != NULL, FALSE);
 	g_return_val_if_fail (priv2->name != NULL, FALSE);
@@ -985,7 +1050,7 @@ soup_address_equal_by_name (gconstpointer addr1, gconstpointer addr2)
 guint
 soup_address_hash_by_ip (gconstpointer addr)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (SOUP_ADDRESS (addr));
 	guint hash;
 
 	g_return_val_if_fail (priv->sockaddr != NULL, 0);
@@ -1023,8 +1088,8 @@ soup_address_hash_by_ip (gconstpointer addr)
 gboolean
 soup_address_equal_by_ip (gconstpointer addr1, gconstpointer addr2)
 {
-	SoupAddressPrivate *priv1 = SOUP_ADDRESS_GET_PRIVATE (addr1);
-	SoupAddressPrivate *priv2 = SOUP_ADDRESS_GET_PRIVATE (addr2);
+	SoupAddressPrivate *priv1 = soup_address_get_instance_private (SOUP_ADDRESS (addr1));
+	SoupAddressPrivate *priv2 = soup_address_get_instance_private (SOUP_ADDRESS (addr2));
 	int size;
 
 	g_return_val_if_fail (priv1->sockaddr != NULL, FALSE);
@@ -1070,7 +1135,7 @@ soup_address_address_enumerator_finalize (GObject *object)
 static GSocketAddress *
 next_address (SoupAddressAddressEnumerator *addr_enum)
 {
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr_enum->addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr_enum->addr);
 	struct sockaddr_storage *ss;
 	int next_addr;
 
@@ -1098,7 +1163,7 @@ soup_address_address_enumerator_next (GSocketAddressEnumerator  *enumerator,
 {
 	SoupAddressAddressEnumerator *addr_enum =
 		SOUP_ADDRESS_ADDRESS_ENUMERATOR (enumerator);
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr_enum->addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr_enum->addr);
 
 	if (!priv->sockaddr) {
 		if (resolve_sync_internal (addr_enum->addr, cancellable, error) != SOUP_STATUS_OK)
@@ -1111,15 +1176,19 @@ soup_address_address_enumerator_next (GSocketAddressEnumerator  *enumerator,
 static void
 got_addresses (SoupAddress *addr, guint status, gpointer user_data)
 {
-	GSimpleAsyncResult *simple = user_data;
+	GTask *task = user_data;
 	GError *error;
 
 	error = g_object_get_data (G_OBJECT (addr), "async-resolved-error");
 	if (error)
-		g_simple_async_result_set_from_error (simple, error);
+		g_task_return_error (task, g_error_copy (error));
+	else {
+		GSocketAddress *addr;
 
-	g_simple_async_result_complete (simple);
-	g_object_unref (simple);
+		addr = next_address (g_task_get_source_object (task));
+		g_task_return_pointer (task, addr, g_object_unref);
+	}
+	g_object_unref (task);
 }
 
 static void
@@ -1130,19 +1199,17 @@ soup_address_address_enumerator_next_async (GSocketAddressEnumerator  *enumerato
 {
 	SoupAddressAddressEnumerator *addr_enum =
 		SOUP_ADDRESS_ADDRESS_ENUMERATOR (enumerator);
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr_enum->addr);
-	GSimpleAsyncResult *simple;
-
-	simple = g_simple_async_result_new (G_OBJECT (enumerator),
-					    callback, user_data,
-					    soup_address_address_enumerator_next_async);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr_enum->addr);
+	GTask *task;
 
+	task = g_task_new (enumerator, cancellable, callback, user_data);
 	if (!priv->sockaddr) {
-		soup_address_resolve_async (addr_enum->addr, NULL, cancellable,
-					    got_addresses, simple);
+		soup_address_resolve_async (addr_enum->addr,
+					    g_main_context_get_thread_default (),
+					    cancellable, got_addresses, task);
 	} else {
-		g_simple_async_result_complete_in_idle (simple);
-		g_object_unref (simple);
+		g_task_return_pointer (task, next_address (addr_enum), g_object_unref);
+		g_object_unref (task);
 	}
 }
 
@@ -1151,14 +1218,7 @@ soup_address_address_enumerator_next_finish (GSocketAddressEnumerator  *enumerat
 					     GAsyncResult              *result,
 					     GError                   **error)
 {
-	SoupAddressAddressEnumerator *addr_enum =
-		SOUP_ADDRESS_ADDRESS_ENUMERATOR (enumerator);
-	GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (result);
-
-	if (g_simple_async_result_propagate_error (simple, error))
-		return NULL;
-	else 
-		return next_address (addr_enum);
+	return g_task_propagate_pointer (G_TASK (result), error);
 }
 
 static void
@@ -1186,9 +1246,9 @@ soup_address_connectable_enumerate (GSocketConnectable *connectable)
 	SoupAddressPrivate *priv;
 
 	addr_enum = g_object_new (SOUP_TYPE_ADDRESS_ADDRESS_ENUMERATOR, NULL);
-	addr_enum->addr = g_object_ref (connectable);
+	addr_enum->addr = g_object_ref (SOUP_ADDRESS (connectable));
 
-	priv = SOUP_ADDRESS_GET_PRIVATE (addr_enum->addr);
+	priv = soup_address_get_instance_private (addr_enum->addr);
 	addr_enum->orig_offset = priv->offset;
 
 	return (GSocketAddressEnumerator *)addr_enum;
@@ -1198,22 +1258,27 @@ static GSocketAddressEnumerator *
 soup_address_connectable_proxy_enumerate (GSocketConnectable *connectable)
 {
 	SoupAddress *addr = SOUP_ADDRESS (connectable);
-	SoupAddressPrivate *priv = SOUP_ADDRESS_GET_PRIVATE (addr);
+	SoupAddressPrivate *priv = soup_address_get_instance_private (addr);
 	GSocketAddressEnumerator *proxy_enum;
-	char *uri;
+	SoupURI *uri;
+	char *uri_string;
 
 	/* We cheerily assume "http" here because you shouldn't be
 	 * using SoupAddress any more if you're not doing HTTP anyway.
 	 */
-	uri = g_strdup_printf ("%s://%s:%u",
-			       priv->protocol ? priv->protocol : "http",
-			       priv->name ? priv->name : soup_address_get_physical (addr),
-			       priv->port);
+	uri = soup_uri_new (NULL);
+	soup_uri_set_scheme (uri, priv->protocol ? priv->protocol : "http");
+	soup_uri_set_host (uri, priv->name ? priv->name : soup_address_get_physical (addr));
+	soup_uri_set_port (uri, priv->port);
+	soup_uri_set_path (uri, "");
+	uri_string = soup_uri_to_string_internal (uri, FALSE, FALSE, TRUE);
+
 	proxy_enum = g_object_new (G_TYPE_PROXY_ADDRESS_ENUMERATOR,
 				   "connectable", connectable,
-				   "uri", uri,
+				   "uri", uri_string,
 				   NULL);
-	g_free (uri);
+	g_free (uri_string);
+	soup_uri_free (uri);
 
 	return proxy_enum;
 }
