@@ -36,6 +36,7 @@ typedef enum {
 	SOUP_MESSAGE_IO_STATE_BODY_START,
 	SOUP_MESSAGE_IO_STATE_BODY,
 	SOUP_MESSAGE_IO_STATE_BODY_DATA,
+	SOUP_MESSAGE_IO_STATE_BODY_FLUSH,
 	SOUP_MESSAGE_IO_STATE_BODY_DONE,
 	SOUP_MESSAGE_IO_STATE_FINISHING,
 	SOUP_MESSAGE_IO_STATE_DONE
@@ -80,6 +81,9 @@ typedef struct {
 	GSource *unpause_source;
 	gboolean paused;
 
+	GCancellable *async_close_wait;
+	GError       *async_close_error;
+
 	SoupMessageGetHeadersFn   get_headers_cb;
 	SoupMessageParseHeadersFn parse_headers_cb;
 	gpointer                  header_data;
@@ -87,8 +91,10 @@ typedef struct {
 	gpointer                  completion_data;
 } SoupMessageIOData;
 	
+static void io_run (SoupMessage *msg, gboolean blocking);
 
 #define RESPONSE_BLOCK_SIZE 8192
+#define HEADER_SIZE_LIMIT (64 * 1024)
 
 void
 soup_message_io_cleanup (SoupMessage *msg)
@@ -120,6 +126,12 @@ soup_message_io_cleanup (SoupMessage *msg)
 	if (io->write_chunk)
 		soup_buffer_free (io->write_chunk);
 
+	if (io->async_close_wait) {
+		g_cancellable_cancel (io->async_close_wait);
+		g_clear_object (&io->async_close_wait);
+	}
+	g_clear_error (&io->async_close_error);
+
 	g_slice_free (SoupMessageIOData, io);
 }
 
@@ -152,7 +164,7 @@ soup_message_io_finished (SoupMessage *msg)
 	SoupMessageIOData *io = priv->io_data;
 	SoupMessageCompletionFn completion_cb;
 	gpointer completion_data;
-	gboolean complete;
+	SoupMessageIOCompletion completion;
 
 	if (!io)
 		return;
@@ -160,16 +172,44 @@ soup_message_io_finished (SoupMessage *msg)
 	completion_cb = io->completion_cb;
 	completion_data = io->completion_data;
 
-	complete = (io->read_state >= SOUP_MESSAGE_IO_STATE_FINISHING &&
-		    io->write_state >= SOUP_MESSAGE_IO_STATE_FINISHING);
+	if ((io->read_state >= SOUP_MESSAGE_IO_STATE_FINISHING &&
+	     io->write_state >= SOUP_MESSAGE_IO_STATE_FINISHING))
+		completion = SOUP_MESSAGE_IO_COMPLETE;
+	else
+		completion = SOUP_MESSAGE_IO_INTERRUPTED;
 
 	g_object_ref (msg);
 	soup_message_io_cleanup (msg);
 	if (completion_cb)
-		completion_cb (msg, complete, completion_data);
+		completion_cb (msg, completion, completion_data);
 	g_object_unref (msg);
 }
 
+GIOStream *
+soup_message_io_steal (SoupMessage *msg)
+{
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+	SoupMessageIOData *io = priv->io_data;
+	SoupMessageCompletionFn completion_cb;
+	gpointer completion_data;
+	GIOStream *iostream;
+
+	if (!io || !io->iostream)
+		return NULL;
+
+	iostream = g_object_ref (io->iostream);
+	completion_cb = io->completion_cb;
+	completion_data = io->completion_data;
+
+	g_object_ref (msg);
+	soup_message_io_cleanup (msg);
+	if (completion_cb)
+		completion_cb (msg, SOUP_MESSAGE_IO_STOLEN, completion_data);
+	g_object_unref (msg);
+
+	return iostream;
+}
+
 static gboolean
 read_headers (SoupMessage *msg, gboolean blocking,
 	      GCancellable *cancellable, GError **error)
@@ -190,6 +230,8 @@ read_headers (SoupMessage *msg, gboolean blocking,
 							    cancellable, error);
 		io->read_header_buf->len = old_len + MAX (nread, 0);
 		if (nread == 0) {
+			if (io->read_header_buf->len > 0)
+				break;
 			soup_message_set_status (msg, SOUP_STATUS_MALFORMED);
 			g_set_error_literal (error, G_IO_ERROR,
 					     G_IO_ERROR_PARTIAL_INPUT,
@@ -202,27 +244,28 @@ read_headers (SoupMessage *msg, gboolean blocking,
 			if (nread == 1 && old_len >= 2 &&
 			    !strncmp ((char *)io->read_header_buf->data +
 				      io->read_header_buf->len - 2,
-				      "\n\n", 2))
+				      "\n\n", 2)) {
+				io->read_header_buf->len--;
 				break;
-			else if (nread == 2 && old_len >= 3 &&
+			} else if (nread == 2 && old_len >= 3 &&
 				 !strncmp ((char *)io->read_header_buf->data +
 					   io->read_header_buf->len - 3,
-					   "\n\r\n", 3))
+					   "\n\r\n", 3)) {
+				io->read_header_buf->len -= 2;
 				break;
+			}
+		}
+
+		if (io->read_header_buf->len > HEADER_SIZE_LIMIT) {
+			soup_message_set_status (msg, SOUP_STATUS_MALFORMED);
+			g_set_error_literal (error, G_IO_ERROR,
+					     G_IO_ERROR_PARTIAL_INPUT,
+					     _("Header too big"));
+			return FALSE;
 		}
 	}
 
-	/* We need to "rewind" io->read_header_buf back one line.
-	 * That SHOULD be two characters (CR LF), but if the
-	 * web server was stupid, it might only be one.
-	 */
-	if (io->read_header_buf->len < 3 ||
-	    io->read_header_buf->data[io->read_header_buf->len - 2] == '\n')
-		io->read_header_buf->len--;
-	else
-		io->read_header_buf->len -= 2;
 	io->read_header_buf->data[io->read_header_buf->len] = '\0';
-
 	return TRUE;
 }
 
@@ -275,6 +318,33 @@ soup_message_setup_body_istream (GInputStream *body_stream,
 	return istream;
 }
 
+static void
+closed_async (GObject      *source,
+	      GAsyncResult *result,
+	      gpointer      user_data)
+{
+	GOutputStream *body_ostream = G_OUTPUT_STREAM (source);
+	SoupMessage *msg = user_data;
+	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (msg);
+	SoupMessageIOData *io = priv->io_data;
+	GCancellable *async_close_wait;
+
+	if (!io || !io->async_close_wait || io->body_ostream != body_ostream) {
+		g_object_unref (msg);
+		return;
+	}
+
+	g_output_stream_close_finish (body_ostream, result, &io->async_close_error);
+	g_clear_object (&io->body_ostream);
+
+	async_close_wait = io->async_close_wait;
+	io->async_close_wait = NULL;
+	g_cancellable_cancel (async_close_wait);
+	g_object_unref (async_close_wait);
+
+	g_object_unref (msg);
+}
+
 /*
  * There are two request/response formats: the basic request/response,
  * possibly with one or more unsolicited informational responses (such
@@ -316,8 +386,28 @@ io_write (SoupMessage *msg, gboolean blocking,
 	SoupBuffer *chunk;
 	gssize nwrote;
 
+	if (io->async_close_error) {
+		g_propagate_error (error, io->async_close_error);
+		io->async_close_error = NULL;
+		return FALSE;
+	} else if (io->async_close_wait) {
+		g_set_error_literal (error, G_IO_ERROR,
+				     G_IO_ERROR_WOULD_BLOCK,
+				     _("Operation would block"));
+		return FALSE;
+	}
+
 	switch (io->write_state) {
 	case SOUP_MESSAGE_IO_STATE_HEADERS:
+		if (io->mode == SOUP_MESSAGE_IO_SERVER &&
+		    io->read_state == SOUP_MESSAGE_IO_STATE_BLOCKING &&
+		    msg->status_code == 0) {
+			/* Client requested "Expect: 100-continue", and
+			 * server did not set an error.
+			 */
+			soup_message_set_status (msg, SOUP_STATUS_CONTINUE);
+		}
+
 		if (!io->write_buf->len) {
 			io->get_headers_cb (msg, io->write_buf,
 					    &io->write_encoding,
@@ -356,6 +446,13 @@ io_write (SoupMessage *msg, gboolean blocking,
 			}
 
 			soup_message_wrote_informational (msg);
+
+			/* If this was "101 Switching Protocols", then
+			 * the server probably stole the connection...
+			 */
+			if (io != priv->io_data)
+				return FALSE;
+
 			soup_message_cleanup_response (msg);
 			break;
 		}
@@ -400,7 +497,7 @@ io_write (SoupMessage *msg, gboolean blocking,
 		if (!io->write_length &&
 		    io->write_encoding != SOUP_ENCODING_EOF &&
 		    io->write_encoding != SOUP_ENCODING_CHUNKED) {
-			io->write_state = SOUP_MESSAGE_IO_STATE_BODY_DONE;
+			io->write_state = SOUP_MESSAGE_IO_STATE_BODY_FLUSH;
 			break;
 		}
 
@@ -412,7 +509,7 @@ io_write (SoupMessage *msg, gboolean blocking,
 				return FALSE;
 			}
 			if (!io->write_chunk->length) {
-				io->write_state = SOUP_MESSAGE_IO_STATE_BODY_DONE;
+				io->write_state = SOUP_MESSAGE_IO_STATE_BODY_FLUSH;
 				break;
 			}
 		}
@@ -442,7 +539,7 @@ io_write (SoupMessage *msg, gboolean blocking,
 	case SOUP_MESSAGE_IO_STATE_BODY_DATA:
 		io->written = 0;
 		if (io->write_chunk->length == 0) {
-			io->write_state = SOUP_MESSAGE_IO_STATE_BODY_DONE;
+			io->write_state = SOUP_MESSAGE_IO_STATE_BODY_FLUSH;
 			break;
 		}
 
@@ -458,13 +555,29 @@ io_write (SoupMessage *msg, gboolean blocking,
 		break;
 
 
-	case SOUP_MESSAGE_IO_STATE_BODY_DONE:
+	case SOUP_MESSAGE_IO_STATE_BODY_FLUSH:
 		if (io->body_ostream) {
-			if (!g_output_stream_close (io->body_ostream, cancellable, error))
-				return FALSE;
-			g_clear_object (&io->body_ostream);
+			if (blocking || io->write_encoding != SOUP_ENCODING_CHUNKED) {
+				if (!g_output_stream_close (io->body_ostream, cancellable, error))
+					return FALSE;
+				g_clear_object (&io->body_ostream);
+			} else {
+				io->async_close_wait = g_cancellable_new ();
+				if (io->async_context)
+					g_main_context_push_thread_default (io->async_context);
+				g_output_stream_close_async (io->body_ostream,
+							     G_PRIORITY_DEFAULT, cancellable,
+							     closed_async, g_object_ref (msg));
+				if (io->async_context)
+					g_main_context_pop_thread_default (io->async_context);
+			}
 		}
 
+		io->write_state = SOUP_MESSAGE_IO_STATE_BODY_DONE;
+		break;
+
+
+	case SOUP_MESSAGE_IO_STATE_BODY_DONE:
 		io->write_state = SOUP_MESSAGE_IO_STATE_FINISHING;
 		soup_message_wrote_body (msg);
 		break;
@@ -546,15 +659,22 @@ io_read (SoupMessage *msg, gboolean blocking,
 			 * bail out here rather than parsing encoding, etc
 			 */
 			soup_message_got_informational (msg);
+
+			/* If this was "101 Switching Protocols", then
+			 * the session may have stolen the connection...
+			 */
+			if (io != priv->io_data)
+				return FALSE;
+
 			soup_message_cleanup_response (msg);
 			break;
 		} else if (io->mode == SOUP_MESSAGE_IO_SERVER &&
 			   soup_message_headers_get_expectations (msg->request_headers) & SOUP_EXPECTATION_CONTINUE) {
-			/* The client requested a Continue response. The
-			 * got_headers handler may change this to something
-			 * else though.
+			/* We must return a status code and response
+			 * headers to the client; either an error to
+			 * be set by a got-headers handler below, or
+			 * else %SOUP_STATUS_CONTINUE otherwise.
 			 */
-			soup_message_set_status (msg, SOUP_STATUS_CONTINUE);
 			io->write_state = SOUP_MESSAGE_IO_STATE_HEADERS;
 			io->read_state = SOUP_MESSAGE_IO_STATE_BLOCKING;
 		} else {
@@ -701,7 +821,7 @@ message_source_check (GSource *source)
 		SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (message_source->msg);
 		SoupMessageIOData *io = priv->io_data;
 
-		if (!io || io->paused)
+		if (io && io->paused)
 			return FALSE;
 		else
 			return TRUE;
@@ -782,6 +902,8 @@ soup_message_io_get_source (SoupMessage *msg, GCancellable *cancellable,
 		base_source = g_timeout_source_new (0);
 	} else if (io->paused) {
 		base_source = NULL;
+	} else if (io->async_close_wait) {
+		base_source = g_cancellable_source_new (io->async_close_wait);
 	} else if (SOUP_MESSAGE_IO_STATE_POLLABLE (io->read_state)) {
 		GPollableInputStream *istream;
 
@@ -857,7 +979,7 @@ io_run_until (SoupMessage *msg, gboolean blocking,
 
 	g_object_ref (msg);
 
-	while (progress && priv->io_data == io && !io->paused &&
+	while (progress && priv->io_data == io && !io->paused && !io->async_close_wait &&
 	       (io->read_state < read_state || io->write_state < write_state)) {
 
 		if (SOUP_MESSAGE_IO_STATE_ACTIVE (io->read_state))
@@ -881,15 +1003,16 @@ io_run_until (SoupMessage *msg, gboolean blocking,
 		g_propagate_error (error, my_error);
 		g_object_unref (msg);
 		return FALSE;
-	} else if (g_cancellable_set_error_if_cancelled (cancellable, error)) {
-		g_object_unref (msg);
-		return FALSE;
 	} else if (priv->io_data != io) {
 		g_set_error_literal (error, G_IO_ERROR,
 				     G_IO_ERROR_CANCELLED,
 				     _("Operation was cancelled"));
 		g_object_unref (msg);
 		return FALSE;
+	} else if (!io->async_close_wait &&
+		   g_cancellable_set_error_if_cancelled (cancellable, error)) {
+		g_object_unref (msg);
+		return FALSE;
 	}
 
 	done = (io->read_state >= read_state &&
@@ -907,8 +1030,6 @@ io_run_until (SoupMessage *msg, gboolean blocking,
 	return done;
 }
 
-static void io_run (SoupMessage *msg, gboolean blocking);
-
 static gboolean
 io_run_ready (SoupMessage *msg, gpointer user_data)
 {
@@ -1200,7 +1321,7 @@ soup_message_io_unpause (SoupMessage *msg)
 
 	if (!io->unpause_source) {
 		io->unpause_source = soup_add_completion_reffed (io->async_context,
-								 io_unpause_internal, msg);
+								 io_unpause_internal, msg, NULL);
 	}
 }
 
