@@ -9,6 +9,7 @@
 #include "config.h"
 #endif
 
+#include <stdlib.h>
 #include <string.h>
 
 #include <glib/gi18n-lib.h>
@@ -16,6 +17,29 @@
 #include "soup.h"
 #include "soup-message-private.h"
 #include "soup-misc-private.h"
+#include "soup-socket-private.h"
+
+static SoupURI *
+parse_connect_authority (const char *req_path)
+{
+	SoupURI *uri;
+	char *fake_uri;
+
+	fake_uri = g_strdup_printf ("http://%s", req_path);
+	uri = soup_uri_new (fake_uri);
+	g_free (fake_uri);
+
+	if (uri->user || uri->password ||
+	    uri->query || uri->fragment ||
+	    !uri->host ||
+	    (uri->port == 0) ||
+	    (strcmp (uri->path, "/") != 0)) {
+		soup_uri_free (uri);
+		return NULL;
+	}
+
+	return uri;
+}
 
 static guint
 parse_request_headers (SoupMessage *msg, char *headers, guint headers_len,
@@ -73,8 +97,11 @@ parse_request_headers (SoupMessage *msg, char *headers, guint headers_len,
 		if (uri)
 			soup_uri_set_path (uri, "*");
 		g_free (url);
+	} else if (msg->method == SOUP_METHOD_CONNECT) {
+		/* Authority */
+		uri = parse_connect_authority (req_path);
 	} else if (*req_path != '/') {
-		/* Must be an absolute URI */
+		/* Absolute URI */
 		uri = soup_uri_new (req_path);
 	} else if (req_host) {
 		url = g_strdup_printf ("%s://%s%s",
@@ -220,6 +247,9 @@ get_response_headers (SoupMessage *msg, GString *headers,
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
 
+	if (msg->status_code == 0)
+		soup_message_set_status (msg, SOUP_STATUS_INTERNAL_SERVER_ERROR);
+
 	handle_partial_get (msg);
 
 	g_string_append_printf (headers, "HTTP/1.%c %d %s\r\n",
@@ -252,17 +282,22 @@ get_response_headers (SoupMessage *msg, GString *headers,
 void
 soup_message_read_request (SoupMessage               *msg,
 			   SoupSocket                *sock,
+			   gboolean                   use_thread_context,
 			   SoupMessageCompletionFn    completion_cb,
 			   gpointer                   user_data)
 {
 	GMainContext *async_context;
 	GIOStream *iostream;
 
-	g_object_get (sock,
-		      SOUP_SOCKET_ASYNC_CONTEXT, &async_context,
-		      NULL);
-	if (!async_context)
-		async_context = g_main_context_ref (g_main_context_default ());
+	if (use_thread_context)
+		async_context = g_main_context_ref_thread_default ();
+	else {
+		g_object_get (sock,
+			      SOUP_SOCKET_ASYNC_CONTEXT, &async_context,
+			      NULL);
+		if (!async_context)
+			async_context = g_main_context_ref (g_main_context_default ());
+	}
 
 	iostream = soup_socket_get_iostream (sock);
 
@@ -271,6 +306,5 @@ soup_message_read_request (SoupMessage               *msg,
 				parse_request_headers,
 				sock,
 				completion_cb, user_data);
-	if (async_context)
-		g_main_context_unref (async_context);
+	g_main_context_unref (async_context);
 }
