@@ -5,10 +5,14 @@
  * Copyright (C) 2000-2003, Ximian, Inc.
  */
 
-#include <ctype.h>
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <string.h>
 
 #include "soup-misc.h"
+#include "soup-misc-private.h"
 
 /**
  * SECTION:soup-misc
@@ -113,6 +117,20 @@ soup_add_idle (GMainContext *async_context,
 	return source;
 }
 
+GSource *
+soup_add_completion_reffed (GMainContext   *async_context,
+			    GSourceFunc     function,
+			    gpointer        data,
+			    GDestroyNotify  dnotify)
+{
+	GSource *source = g_idle_source_new ();
+
+	g_source_set_priority (source, G_PRIORITY_DEFAULT);
+	g_source_set_callback (source, function, data, dnotify);
+	g_source_attach (source, async_context);
+	return source;
+}
+
 /**
  * soup_add_completion: (skip)
  * @async_context: (allow-none): the #GMainContext to dispatch the I/O
@@ -133,10 +151,9 @@ GSource *
 soup_add_completion (GMainContext *async_context,
 	             GSourceFunc function, gpointer data)
 {
-	GSource *source = g_idle_source_new ();
-	g_source_set_priority (source, G_PRIORITY_DEFAULT);
-	g_source_set_callback (source, function, data, NULL);
-	g_source_attach (source, async_context);
+	GSource *source;
+
+	source = soup_add_completion_reffed (async_context, function, data, NULL);
 	g_source_unref (source);
 	return source;
 }
@@ -225,3 +242,38 @@ const char soup_char_attributes[] = {
 	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
 	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
 };
+
+/**
+ * soup_host_matches_host
+ * @host: a URI
+ * @compare_with: a URI
+ *
+ * Checks if the @host and @compare_with exactly match or prefixed with a dot.
+ *
+ * Return value: %TRUE if the hosts match, %FALSE otherwise
+ *
+ * Since: 2.54
+ **/
+gboolean
+soup_host_matches_host (const gchar *host, const gchar *compare_with)
+{
+	char *match;
+	int dlen;
+
+	g_return_val_if_fail (host != NULL, FALSE);
+	g_return_val_if_fail (compare_with != NULL, FALSE);
+
+	if (!g_ascii_strcasecmp (host, compare_with))
+		return TRUE;
+	if (*host != '.')
+		return FALSE;
+	if (!g_ascii_strcasecmp (host + 1, compare_with))
+		return TRUE;
+	dlen = strlen (host);
+	while ((match = strstr (compare_with, host))) {
+		if (!match[dlen])
+			return TRUE;
+		compare_with = match + 1;
+	}
+	return FALSE;
+}
